# Statistics functions

adjustedPSidak <- function(p,n){
  1-(1-p)^n
}

adjustedPHolm <- function(pArray){
  adjustedPs <- rep(0,length(pArray))
  m <- length(pArray)
  sortedArr <- sort(pArray)
  pOrder <- order(pArray)
  for (k in 1:m){
    adjPrk <- (m-k+1)*sortedArr[k]
    adjustedPs[pOrder[k]] <- max(c(adjustedPs[pOrder[1:k]],adjPrk))
  }
  adjustedPs
}

# Westfall and Young (1993); adjusted Pri = max(k=1..i){ Pr( min(l=k..m)Prl <= prk | Hm) }
# samples are rows, observations are cols
adjustedPWestfallYoung <- function(pArray,nullPMatrix,lessThan=F){
  # Make sure each row in nullPMatrix is sorted
  numSamples <- dim(nullPMatrix)[1]
  for (i in 1:numSamples){
    nullPMatrix[i,] <- sort(nullPMatrix[i,])
  }

  m <- length(pArray)
  sortedPs <- sort(pArray)
  pOrder <- order(pArray)
  adjustedPValues <- rep(0,m)
  names(adjustedPValues) <- names(pArray)
  for (i in 1:m){
    adjustedPValue <- sum(nullPMatrix[,i] <= sortedPs[i])/numSamples
    if (lessThan & adjustedPValue < 1){
      adjustedPValue <- adjustedPValue + 1/numSamples
    }
    adjustedPValues[pOrder[i]] <- max(adjustedPValue,adjustedPValues[pOrder[1:i]])
  }
  adjustedPValues
}

adjustedPValues <- function(data1,data2,test=wilcox.test){
  numTests <- ncol(data1)
  pVals <- rep(1,numTests)
  for (i in 1:numTests){
    wtest <- wilcox.test(data1[[i]],data2[[i]])
    rawPVal <- wtest$p.value
    rawEVal <- -log(1-rawPVal)
    adjEVal <- rawEVal*25
    adjPVal <- 1-exp(-adjEVal)
    pVals[i] <- adjPVal
  }
  return(pVals)
}

skewness <- function(x,na.rm=T) {
  m3 <- mean((x-mean(x,na.rm=na.rm))^3,na.rm=na.rm)
  skew <- m3/(sd(x,na.rm=na.rm)^3)
  return(skew)
}

kurtosis <- function(x,na.rm=T){
  return(standardizedMoment(x,4,na.rm=na.rm)-3)
}

moment <- function(x,n=2,na.rm=T){
  return(sum((x-mean(x,na.rm=na.rm))^n,na.rm=na.rm)/(length(x)-1))
}

standardizedMoment <- function(x,n=2,na.rm=T){
  return(moment(x,n,na.rm=na.rm)/var(x,na.rm=na.rm)^(n/2))
}

percentile <- function(value,distribution){
  return(max((sum(distribution<value)-.5),0)/length(distribution))
}

percentileRank <- function(value,distribution,mid=T){
  if (mid){
    if (sum(value==distribution) == 0){
      equal <- 0
    }else{
      equal <- (1+sum(value==distribution))/2
    }
  }else{
    equal <- sum(value==distribution)
  }
  return(max((sum(distribution<value)+equal-.5),0)/length(distribution))
}

BIC.MDSStress <- function(distanceData,stress=NULL,numDimensions=1:10,alpha=1){
  if (is.null(stress)){
    stress <- as.numeric(mclapply(numDimensions,function(m,distanceData){
      return(isoMDS(distanceData,k=m)$stress)
    },distanceData))
    names(stress) <- numDimensions
  }
  n <- round(.5 + sqrt(.25 + 2*length(distanceData)),0)
  return(alpha*stress + numDimensions*n*log(n*(n-1)/2))
}

BIC.MDS <- function(distanceData,numDimensions=1:10,alpha=1,measure=c("Stress","Error","NormError")){
  measure <- match.arg(measure)
  bic <- rep(0,length(numDimensions))
  names(bic) <- numDimensions
  require(parallel)
  numCores <- detectCores()
  if (!is.na(numCores)){
    bic <- as.numeric(mclapply(numDimensions,function(m,distanceData,alpha,measure){
      if (class(distanceData) != "dist"){
        n <- nrow(distanceData)
        distanceData <- as.dist(distanceData)
      }else{
        n <- round(.5 + sqrt(.25 + 2*length(distanceData)),0)
      }
      logNumDistances <- log(n*(n-1)/2)
      mdsData <- isoMDS(distanceData,k=m)
      
      if (measure == "Stress"){
        error <- mdsData$stress
      }else if (measure == "Error"){
        mdsDistances <- dist(mdsData$points)
        error <- sum((distanceData - mdsDistances)^2)
      }else if (measure == "NormError"){
        mdsDistances <- dist(mdsData$points)
        error <- sum((scale(distanceData) - scale(mdsDistances))^2)
      }
      return(alpha*error + m*n*logNumDistances)
      },distanceData,alpha,measure,mc.cores=numCores))
  }
  else{
    if (class(distanceData) != "dist"){
      distanceData <- as.dist(distanceData)
    }
    logNumDistances <- log(n*(n-1)/2)
    for (i in 1:length(numDimensions)){
      m <- numDimensions[i]
      mdsData <- isoMDS(distanceData,k=m)
      mdsDistances <- dist(mdsData,mdsData$points)
      if (measure == "Stress"){
        error <- mdsData$stress
      }else if (measure == "Error"){
        mdsDistances <- dist(mdsData$points)
        error <- sum((distanceData - mdsDistances)^2)
      }else if (measure == "NormError"){
        mdsDistances <- dist(mdsData$points)
        error <- sum((scale(distanceData) - scale(mdsDistances))^2)
      }
      bic[i] <- alpha*error + m*n*logNumDistances
    }
  }
  return(bic)
}

MDSSumSquareError <- function(distanceData,numDimensions=1:10,percent=F,normalize=F){
  sse <- rep(0,length(numDimensions))
  names(sse) <- numDimensions
  # summary(dist(mdsData[,1:m])/as.dist(distanceData))
  if (class(distanceData) != "dist"){
    distanceData <- as.dist(distanceData)
  }
  require(parallel)
  numCores <- detectCores()
  if (!is.na(numCores)){
    sse <- as.numeric(mclapply(numDimensions,function(m,distanceData,alpha,percent,normalize){
      if (class(distanceData) != "dist"){
        distanceData <- as.dist(distanceData)
        n <- nrow(distanceData)
      }else{
        n <- round(.5 + sqrt(.25 + 2*length(distanceData)),0)
      }
      logNumDistances <- log(n*(n-1)/2)
      mdsData <- isoMDS(distanceData,k=m)
      mdsDistances <- dist(mdsData$points)
      if (normalize){
        error <- sum((scale(distanceData) - scale(mdsDistances))^2)
      }else{
        error <- sum((distanceData - mdsDistances)^2)
      }
      if (percent){
        sse <- error/length(error)
      }else{
        sse <- error
      }
      return(sse)
    },distanceData,alpha,percent,normalize,mc.cores=numCores))
  }else{
    for (i in 1:length(numDimensions)){
      m <- numDimensions[i]
      mdsData <- isoMDS(distanceData,k=m)
      mdsDistances <- dist(mdsData$points)
      if (normalize){
        error <- sum((scale(distanceData) - scale(mdsDistances))^2)
      }else{
        error <- sum((distanceData - mdsDistances)^2)
      }
      if (percent){
        sse[i] <- error/length(error)
      }else{
        sse[i] <- error
      }
    }
  }
  return(sse)
}

# create functions to get the lower and upper bounds of the error bars
stderr <- function(x){sqrt(var(x,na.rm=TRUE)/length(na.omit(x)))}
lowsd <- function(x){return(mean(x)-stderr(x))}
highsd <- function(x){return(mean(x)+stderr(x))}

stderrOfSD <- function(x){
  n<-length(x)
  return(sd(x)*sqrt((n-1 - 2*gamma(n/2)^2/gamma((n-1)/2)^2)/(n-1)))
}
lowsesd <- function(x){return(sd(x)-stderrOfSD(x))}
highsesd <- function(x){return(sd(x)+stderrOfSD(x))}
