#############################################################################################
#
# hwdRealign.r Originally for recalculating height, width, and depth for each arbor based
#    on reorienting using PCA. Dendrites (including apical) are oriented given the entire
#    dendritic arbor. Axons are oriented just with the axonal arbor.
#
#    Additionally, code at the end enables joining these and other measures with 
#    NMO morphometrics.
#
#############################################################################################

require(ade4)
require(parallel)

homeDir <- ""

# Set up rotation variables
xRotatePos90 <- matrix(c(1, 0, 0,  0, 0,-1,  0, 1, 0),3,3)
xRotateNeg90 <- matrix(c(1, 0, 0,  0, 0, 1,  0,-1, 0),3,3)
yRotatePos90 <- matrix(c(0, 0, 1,  0, 1, 0, -1, 0, 0),3,3)
yRotateNeg90 <- matrix(c(0, 0,-1,  0, 1, 0,  1, 0, 0),3,3)
zRotatePos90 <- matrix(c(0,-1, 0,  1, 0, 0,  0, 0, 1),3,3)
zRotateNeg90 <- matrix(c(0, 1, 0, -1, 0, 0,  0, 0, 1),3,3)
zRotate180   <- matrix(c(-1,0, 0,  0,-1, 0,  0, 0, 1),3,3)

neuronNamePattern <- ".*\\/([^\\/\\.]+)(\\.CNG)?\\.swc"

pcaAlign <- function(neuronFiles,saveRotatedDir=NULL,overwrite=T,calcHWD=T,neuriteNames=NULL,checkPlotDir=NULL,plotDepth=F,somaDistRatioThreshold=0.34,zDist0sThreshold=0.3,runArbors=list(2,3,4),rotateOnArbors=list(2,c(3,4),c(3,4)),centerToSomaAlign=T,somaCenter=F,updatedArbors=c(),numCores=detectCores()){
  HWDData <- NULL

  if (!is.na(numCores) && numCores > 1){
    results <- mclapply(neuronFiles,pcaAlignSub,saveRotatedDir,overwrite,calcHWD,neuriteNames,checkPlotDir,plotDepth,somaDistRatioThreshold,zDist0sThreshold,runArbors,rotateOnArbors,centerToSomaAlign,somaCenter,updatedArbors,mc.cores=numCores)
    if (calcHWD){
      for (i in 1:length(results)){
        if (is.null(HWDData)){
          HWDData <- results[[i]]
        }else if (!is.null(results[[i]]) && ncol(results[[i]])==3){
          HWDData <- rbind(HWDData,results[[i]])
        }else{
          print(paste("Problem with ",i,"- result value:",results[[i]]))
        }
      }
    }
  }else{
    for (i in 1:length(neuronFiles)){
      newData <- pcaAlignSub(neuronFiles[i],saveRotatedDir,overwrite,calcHWD,neuriteNames,checkPlotDir,plotDepth,somaDistRatioThreshold,zDist0sThreshold,runArbors,rotateOnArbors,centerToSomaAlign,somaCenter,updatedArbors)
      if (calcHWD){
        if (is.null(HWDData)){
          HWDData <- newData
        }else if (!is.null(newData)&& ncol(newData)==3){
          HWDData <- rbind(HWDData,newData)
        }
      }
    }
  }
  return(HWDData)
}

# Might not do well if the rotation basis arbor(s) is not included in the measurement
pcaAlignSub <- function(neuronFilename,saveRotatedDir=NULL,overwrite=T,calcHWD=T,neuriteNames=NULL,checkPlotDir=NULL,plotDepth=F,somaDistRatioThreshold=0.34,zDist0sThreshold=0.3,runArbors=list(2,3,4),rotateOnArbors=list(2,c(3,4),c(3,4)),centerToSomaAlign=T,somaCenter=F,updatedArbors=c()){
  newData <- NULL
  checkPlot <- !is.null(checkPlotDir)

  # Read in swc data
  filenameParts <- strsplit(neuronFilename,"/")[[1]]
  neuronName <- gsub("(.+)\\.CNG\\.swc","\\1",filenameParts[length(filenameParts)],perl=T)
  
  rotatedFilename <- paste(saveRotatedDir,"/",neuronName,".swc",sep="")
  somethingToDo <- !is.null(saveRotatedDir) && (overwrite || !file.exists(rotatedFilename)) || calcHWD || checkPlot
  
  if (somethingToDo && length(grep("\\.swc",neuronFilename))>0){
    print(paste("Running",neuronName))
    
    neuronSwc <- read.table(neuronFilename,stringsAsFactor=F,fill=T)
    neuronSwc <- neuronSwc[,!apply(neuronSwc,2,function(x){all(is.na(x))})]
    colnames(neuronSwc) <- c("id","type","x","y","z","radius","parent")
    for (colInd in 3:5){
      if (!is.numeric(neuronSwc[,colInd])){
        neuronSwc[,colInd] <- as.numeric(gsub(',','',neuronSwc[,colInd]))
      }
    }
    neuronSwc$parent <- as.character(neuronSwc$parent)
    
    # Center to the soma (id 1)
    neuronSwc$x <- neuronSwc$x - neuronSwc$x[1]
    neuronSwc$y <- neuronSwc$y - neuronSwc$y[1]
    neuronSwc$z <- neuronSwc$z - neuronSwc$z[1]
    
    # Calculate compartment lengths
    rownames(neuronSwc) <- neuronSwc$id
    nonRootSegs <- as.numeric(neuronSwc$parent) > 0
    nonRootSwc <- subset(neuronSwc,nonRootSegs)
    compartmentLengths <- sqrt((nonRootSwc$x-nonRootSwc[nonRootSwc$parent,"x"])^2 + 
                                 (nonRootSwc$y-nonRootSwc[nonRootSwc$parent,"y"])^2 +
                                 (nonRootSwc$z-nonRootSwc[nonRootSwc$parent,"z"])^2)
    neuronSwc[nonRootSegs,"compartmentLength"] <- compartmentLengths
    neuronSwc[which(is.na(neuronSwc$compartmentLength)),"compartmentLength"] <- 0
    
    # Calculate compartment centers
    centerPosNames <- c("centerX","centerY","centerZ")    
    neuronSwc[nonRootSegs,"centerX"] <- neuronSwc[nonRootSwc$parent,"x"]+(nonRootSwc$x-neuronSwc[nonRootSwc$parent,"x"])/2
    neuronSwc[nonRootSegs,"centerY"] <- neuronSwc[nonRootSwc$parent,"y"]+(nonRootSwc$y-neuronSwc[nonRootSwc$parent,"y"])/2
    neuronSwc[nonRootSegs,"centerZ"] <- neuronSwc[nonRootSwc$parent,"z"]+(nonRootSwc$z-neuronSwc[nonRootSwc$parent,"z"])/2
    
    nonRootSwc <- subset(neuronSwc,nonRootSegs)
    #unique(neuronSwc$type)
    
    #for (myType in 3:4){
    #    for (myType in c(3)){
    i <- 0
    for (myType in runArbors){
      i <- i+1
      if (is.list(rotateOnArbors)){
        rotateArb <- rotateOnArbors[[i]]
      }else{
        rotateArb <- rotateOnArbors[i]
      }
    
      if (sum(c(2:4)%in%myType)==3){ # All arbor types processed together
        neuriteName <- neuronName
        neuritesOfInterest <- nonRootSwc
        pos <- integer(0)
      }else if (sum(c(3:4)%in%myType)==2){ # Dendrites (basal and apical) processed together
        neuriteName <- paste(neuronName,"Dendrites",sep="-")
        neuritesOfInterest <- subset(nonRootSwc,type%in%myType)
        #pos <- grep(paste("^",gsub("([\\(\\)])","\\\\\\1",neuronName),"-Dendrites",sep=""),rownames(Morphometrics))
      }else if (length(myType) == 1){ # Only one arbor type being processed
        if (!is.null(neuriteNames)){
          typePatterns <- c("","Axon","(Basal)?Dendrite","ApicalDendrite")
          pos <- grep(paste("^",gsub("([\\(\\)])","\\\\\\1",neuronName),"-",typePatterns[myType],sep=""),neuriteNames)
          if (length(pos) > 0){
            neuriteName <- neuriteNames[pos[1]]
          }else{
            # If this neurite is not in the list of neurite names, just skip it
            next
          }
        }else{
          typeNames <- c("","Axon","Dendrite","ApicalDendrite")
          neuriteName <- paste(neuronName,typeNames[myType],sep="-")
        }
        neuritesOfInterest <- subset(nonRootSwc,type%in%myType)
      }else{
        neuriteName <- NULL
        pos <- integer(0)
        warning(paste("Invalid runArbors element ",myType,". Elements must arrays containing (2,3,4), (3,4), or any of 2,3,4 individually.",sep=""))
        next
      }
      
      if (length(rotateArb) == 0){
        neuriteOfType <- nonRootSwc
      }else{
        neuriteOfType <- subset(nonRootSwc,type%in%rotateArb)
      }

      # Not worth running this case if not enough data
      if (nrow(neuriteOfType) < 5 || nrow(neuritesOfInterest) < 5 || neuriteName%in%updatedArbors){
        next
      }
      
      if (calcHWD || !is.null(saveRotatedDir) || checkPlot){
         # If this neurite is in the Morphometrics data and the current type is in the file
         #if (length(pos) > 0 && ){
        #print(paste("Type:",myType))
        
        ####
        # Check whether Z should automatically be depth (if it is sufficiently discrete)
        zDists <- neuriteOfType$z-neuronSwc[neuriteOfType$parent,"z"]
        
        # Run PCA
        zIncluded <- T
        if (mean(zDists==0) >= zDist0sThreshold){ 
          # If Z has insufficient resolution, ignore in PCA
          zIncluded <- F
          if (somaCenter){
            arborPCs <- dudi.pca(neuriteOfType[,centerPosNames[1:2]],center=F,scale=F,row.w=neuriteOfType$compartmentLength,scannf=F,nf=2)
          }else{
            arborPCs <- dudi.pca(neuriteOfType[,centerPosNames[1:2]],center=T,scale=F,row.w=neuriteOfType$compartmentLength,scannf=F,nf=2)
          }
          rotation <- as.matrix(arborPCs$c1)
          # Add Z
          rotation <- cbind(rotation,CS3=c(0,0))
          rotation <- as.matrix(rbind(rotation,data.frame(CS1=0,CS2=0,CS3=1)))
        }else{
          if (somaCenter){
            arborPCs <- dudi.pca(neuriteOfType[,centerPosNames],center=F,scale=F,row.w=neuriteOfType$compartmentLength,scannf=F,nf=3)
          }else{
            arborPCs <- dudi.pca(neuriteOfType[,centerPosNames],center=T,scale=F,row.w=neuriteOfType$compartmentLength,scannf=F,nf=3)
            #arborPCs3 <- prcomp(neuriteOfType[,centerPosNames],center=T,scale.=F)
          }
          rotation <- as.matrix(arborPCs$c1)
          if (ncol(rotation)==2){
            rotation <- cbind(rotation,CS3=c(0,0,1))
          }
        }
        if (nrow(rotation) != 3 || ncol(rotation) != 3){
          print(paste("Rotation matrix for",neuriteName,"dims:",paste(dim(rotation),collapse=",")))
        }
        if (det(rotation) == -1){
          rotation[,3] <- -rotation[,3]
        }
        
        depthZero <- F
        if (ncol(rotation) == 2){ # If range in Z is 0
          depthZero <- T
          rotation <- cbind(rotation,c(0,0,0))
        }
        # Rotate the points
        rotSegmentCenters <- as.matrix(neuriteOfType[,centerPosNames])%*%rotation
        pcNames <- c("PC1","PC2","PC3")
        pcPoints <- as.matrix(neuronSwc[,3:5])%*%rotation
        colnames(pcPoints) <- pcNames
        
        # Get the center of mass (COM) and avg dist of points from COM
        centerOfMass <- apply(t(rotSegmentCenters)%*%neuriteOfType$compartmentLength,1,sum)/sum(neuriteOfType$compartmentLength)
        columns <- 1:3
        if (!zIncluded) {
          columns <- 1:2
        }
        somaToCOMDist <- magnitude(centerOfMass[columns])
        avgDist <- sum(apply(rotSegmentCenters[,columns],1,magnitude)*neuriteOfType$compartmentLength)/sum(neuriteOfType$compartmentLength)
        somaDistRatio <- somaToCOMDist/avgDist
        
        print(paste("Neurite: ",myType,"; Soma Dist: ",round(somaToCOMDist,1),"; AvgDist: ",round(avgDist,1),"; Ratio: ",round(somaDistRatio,3),sep=""))
        
        # Get the  the unit vector from soma to COM
        comVect <- centerOfMass/(sqrt(centerOfMass%*%centerOfMass))
        if (centerToSomaAlign && somaDistRatio > somaDistRatioThreshold){
          # Determine the axis most closely aligned with the soma to COM vector
          heightAxis <- which(abs(comVect)==max(abs(comVect)))
          
          if (heightAxis == 1){
            # Switch X(PC1->Height) and Y(PC2->Width) by rotating around Z
            if (comVect[1] < 0){
              pcPoints <- pcPoints%*%zRotateNeg90
            }else{
              pcPoints <- pcPoints%*%zRotatePos90
            }
          } else if (heightAxis == 2){
            if (comVect[2] < 0){
              # Reverse Y(PC2->Height) by rotating around Z
              pcPoints <- pcPoints%*%zRotate180
            }
          } else if (heightAxis == 3){
            if (comVect[3] < 0){
              # Reverse Z(PC3->Height) and Y by rotating around X(PC1, stays Width)
              pcPoints <- pcPoints%*%xRotatePos90
            }else{
              pcPoints <- pcPoints%*%xRotateNeg90
            }
            pcPoints <- pcPoints%*%yRotatePos90
          }
        }else{
          # Use the first PC as the Height axis (put in second column)
          if (comVect[1] < 0){
            pcPoints <- pcPoints%*%zRotateNeg90
          }else{
            pcPoints <- pcPoints%*%zRotatePos90
          }
        }
        
        hwdNames <- c("Width","Height","Depth")
        colnames(pcPoints) <- hwdNames
        neuronHwdData <- cbind(neuronSwc,as.data.frame(pcPoints))
        
        #hwdData <- subset(neuronHwdData,type==myType) # TYPE-SPECIFIC
        hwdData <- subset(neuronHwdData,type%in%c(1,3,4))
        centerOfMass <- apply(t(as.matrix(hwdData[,hwdNames]))%*%hwdData$compartmentLength,1,sum)/sum(hwdData$compartmentLength)
        
        if (!is.null(saveRotatedDir) && (overwrite || !file.exists(rotatedFilename))){
          outData <- neuronSwc[,1:7]
          outData$y <- round(neuronHwdData$Height,3)
          outData$x <- round(neuronHwdData$Width,3)
          outData$z <- round(neuronHwdData$Depth,3)
          write.table(outData,rotatedFilename,col.names=F,row.names=F,quote=F)
        }
        
        # Check plot
        if (checkPlot && !zIncluded){
          # Plot along principale component axes
          #pcPoints <- as.matrix(neuronSwc[,3:5])%*%arborPCs$rotation
          #pcPoints <- arborPCs$x
          #pcPoints <- as.matrix(neuriteOfType[,3:5])%*%arborPCs$rotation
          #pcPoints <- as.matrix(neuriteOfType[,3:5])%*%as.matrix(arborPCs$c1)
          #par(ask=T)
          axisLim <- max(c(diff(range(hwdData$Height)),diff(range(hwdData$Width)),diff(range(hwdData$Depth))))
          
          xLim <- mean(range(hwdData$Width))+c(-1,1)*axisLim/2
          yLim <- mean(range(hwdData$Height))+c(-1,1)*axisLim/2
          png(paste(checkPlotDir,"/",neuronName,"_WH.png",sep=""))
          #plot(hwdData$Width,hwdData$Height,col=myColors[hwdData$type],cex=.2,xlim=xLim,ylim=yLim,main=neuriteName,xlab="Width",ylab="Height")
          plot(hwdData$Width,hwdData$Height,col=myColors[hwdData$type],cex=.2,xlim=xLim,ylim=yLim,main=paste(neuronName,"\n(Soma Distance Ratio: ",round(somaDistRatio,3),")\n(Z-Dist 0s Proportion: ",round(mean(zDists==0),2),")",sep=""),xlab="Width",ylab="Height")
          points(0,0,col=1,pch=21,bg=1,cex=1.2)
          points(centerOfMass[1],centerOfMass[2],col=1,pch=21,bg=2,cex=1.2)
          dev.off()
          
          # Depth-Height
          png(paste(checkPlotDir,"/",neuronName,"_DH.png",sep=""))
          xLim <- mean(range(hwdData$Depth))+c(-1,1)*axisLim/2
          yLim <- mean(range(hwdData$Height))+c(-1,1)*axisLim/2
          plot(hwdData$Depth,hwdData$Height,col=myColors[hwdData$type],cex=.2,xlim=xLim,ylim=yLim,main=neuriteName,xlab="Depth",ylab="Height")
          dev.off()
          
          if (!depthZero && plotDepth){
            xLim <- mean(range(hwdData$Depth))+c(-1,1)*axisLim/2
            yLim <- mean(range(hwdData$Height))+c(-1,1)*axisLim/2
            plot(hwdData$Depth,hwdData$Height,col=myColors[hwdData$type],cex=.2,xlim=xLim,ylim=yLim,main=neuriteName)
            points(0,0,col=1)
            points(centerOfMass[3],centerOfMass[2],col=2)
            
            xLim <- mean(range(hwdData$Depth))+c(-1,1)*axisLim/2
            yLim <- mean(range(hwdData$Width))+c(-1,1)*axisLim/2
            plot(hwdData$Depth,hwdData$Width,col=myColors[hwdData$type],cex=.2,xlim=xLim,ylim=yLim,main=neuriteName)
            points(0,0,col=1)
            points(centerOfMass[3],centerOfMass[1],col=2)
          }
        }
        
        #####################  
        # Calculate width, height, and depth
        if (calcHWD){
          # Using this to get H/W/D on the entire neuron including axon - for Sridevi to compare
          myType <- 0
          
          height <- calculate95Perc(neuronHwdData,myType,"Height")
          #Morphometrics[pos,"Height"] <- height
          
          width <- calculate95Perc(neuronHwdData,myType,"Width")
          #Morphometrics[pos,"Width"] <- width
          
          if (depthZero){
            depth <- 0
          }else{
            depth <- calculate95Perc(neuronHwdData,myType,"Depth")
          }
          #Morphometrics[pos,"Depth"] <- depth
          
          if (is.null(newData)){
            newData <- data.frame(Height=height,Width=width,Depth=depth)
            rownames(newData) <- neuriteName
          }else{
            newDataTmp <- data.frame(Height=height,Width=width,Depth=depth)
            rownames(newDataTmp) <- neuriteName
            newData <- rbind(newData,newDataTmp)
          }
          
          updatedArbors <- append(updatedArbors,neuriteName)
        }
      }
    }
  }
  return(newData)
}

vcrossp <- function(a, b){
  result <- t(c(0,0,0))
  result[1] <- a[2] * b[3] - a[3] * b[2] 
  result[2] <- a[3] * b[1] - a[1] * b[3] 
  result[3] <- a[1] * b[2] - a[2] * b[1] 
  return(result)
}

calculate95Perc <- function(myData, myType, column){
  if (myType > 0){
    typeData <- subset(myData,type==myType&parent!="-1")
  }
  else{
    typeData <- subset(myData,parent!="-1")
  }
  myDataOrder <- order(typeData[,column])
  typeData <- typeData[myDataOrder,]
  lengthProportion <- typeData$compartmentLength/sum(typeData$compartmentLength)
  cumSumPos <- c(0,cumsum(lengthProportion))
  cumSumNeg <- 1-cumSumPos
  perc95Dists <- c()
  i <- 1
  while (cumSumPos[i] <= .05){
    matchingInd <- which(cumSumPos[i]+cumSumNeg >= .05)
    matchingInd <- matchingInd[length(matchingInd)]
    if (i==0){
      posI <- myData[i,column]
    }else{
      posI <- myData[typeData[i,"parent"],column]
    }
    perc95Dists <- append(perc95Dists,abs(posI-typeData[matchingInd,column]))
    if (is.na(perc95Dists[length(perc95Dists)])){
      break
    }
    i <- i+1
  }
  return(min(perc95Dists))
}

######################
# Code to update morphometrics with Height, Width, Depth calculations
swcNMODir <- "/home/tgillett/swcFiles/NM/"
#MorphometricsHWD <- read.csv(paste(clusterDir,"/NM-Morphometrics_HWD.csv",sep=""),row.names=1)
MorphometricsHWD <- read.csv(paste(clusterDir,"/NM-Morphometrics_HWD_070113.csv",sep=""),row.names=1)
Morphometrics <- read.csv(paste(clusterDir,"/NM-Morphometrics.csv",sep=""),row.names=1)
Morphometrics$Height <- rep(NA,nrow(Morphometrics))
Morphometrics$Width <- rep(NA,nrow(Morphometrics))
Morphometrics$Depth <- rep(NA,nrow(Morphometrics))
sharedMetricSeqs <- intersect(rownames(Morphometrics),rownames(MorphometricsHWD))

# For running only neurons without HWD already calculated
#sharedMetricSeqs <- sharedMetricSeqs[!is.na(MorphometricsHWD[sharedMetricSeqs,"Height"])]
Morphometrics[sharedMetricSeqs,c("Height","Width","Depth")] <- MorphometricsHWD[sharedMetricSeqs,c("Height","Width","Depth")]
allFiles <- paste(swcNMODir,list.files(swcNMODir,recursive=T),sep="")
neuritesToRun <- rownames(Morphometrics)[is.na(Morphometrics$Height)]

# For running all neurons
allFiles <- paste(swcNMODir,list.files(swcNMODir,recursive=T),sep="")
neuritesToRun <- sharedMetricSeqs

## Continue for both cases
# Reduce allFiles list to those containing neurites we want to process
neuronListP1 <- gsub(".*\\/([^\\.\\/]+)\\.CNG\\.swc","\\1",allFiles) # get neuron names from allFiles
neuronListP2 <- gsub("-(ApicalDendrite|BasalDendrite|Dendrite|Axon)","",neuritesToRun) # get neuron names from neurites list
allFiles <- allFiles[neuronListP1%in%neuronListP2]

HWDData <- pcaAlign(allFiles,neuriteNames=neuritesToRun,zDist0sThreshold=1,numCores=8) # Doesn't check Z for resolution
HWDDataCheck <- HWDData[intersect(rownames(HWDData),rownames(Morphometrics)),]
Morphometrics[rownames(HWDDataCheck),"Height"] <- HWDDataCheck$Height
Morphometrics[rownames(HWDDataCheck),"Width"] <- HWDDataCheck$Width
Morphometrics[rownames(HWDDataCheck),"Depth"] <- HWDDataCheck$Depth

if (!"Caulescence"%in%colnames(Morphometrics)){
  Caulescence <- read.csv(paste(clusterDir,"/NM-Caulescence2015.01.12.csv",sep=""),row.names=1)
  Morphometrics$Caulescence <- rep(NA,nrow(Morphometrics))
  Morphometrics[rownames(Morphometrics),"Caulescence"] <- Caulescence[rownames(Morphometrics),"Caulescence"]
}

write.csv(Morphometrics,paste(clusterDir,"/NM-Morphometrics_HWD.csv",sep=""))

######################
